*! version 2.1.6  29feb2016  Ben Jann

program texdoc
    version 9.2
    gettoken subcmd 0 : 0, parse(", ")
    local length = length(`"`subcmd'"')
    if `"`subcmd'"'==substr("write", 1, max(`length', 1)) {
        if `"${TeXdoc_docname}"'=="" {
            di as txt "(texdoc not initialized; nothing to do)"
            exit
        }
        _texdoc write`macval(0)'
        exit
    }
    if `"`subcmd'"'==substr("init", 1, max(`length', 1)) {
        texdoc_init`macval(0)'
        exit
    }
    if `"`subcmd'"'==substr("close", 1, max(`length', 1)) {
        texdoc_close`macval(0)'
        exit
    }
    if `"`subcmd'"'==substr("append", 1, max(`length', 1)) {
        texdoc_append`macval(0)'
        exit
    }
    if `"`subcmd'"'==substr("stlog", 1, max(`length', 1)) {
        local caller : di _caller()
        version `caller': texdoc_stlog`macval(0)'
        exit
    }
    if `"`subcmd'"'==substr("graph", 1, max(`length', 2)) {
        texdoc_graph`macval(0)'
        exit
    }
    if `"`subcmd'"'=="strip" {
        texdoc_strip`macval(0)'
        exit
    }
    if `"`subcmd'"'=="do" {
        local caller : di _caller()
        local do_globals cmdlogbuf replace append logdir logdir2 noprefix ///
            prefix prefix2 stpath stpath2 nodo nolog cmdlog verbatim hardcode ///
            nokeep custom cmdstrip lbstrip noltrim gropts grdir
        local init_globals nodo nolog cmdlog verbatim hardcode nokeep custom ///
            cmdstrip lbstrip noltrim gropts grdir stpath prefix logdir ///
            path0 path basename docname0 docname
        local nested = `"$TeXdoc_dofile"'!=""
        if `nested' { // backup current settings
            local do_dofile `"$TeXdoc_dofile"'
            foreach g of local do_globals {
                mata: st_local("do_`g'", st_global("TeXdoc_do_`g'"))
            }
            local init_stcounter `"$TeXdoc_stcounter"'
            foreach g of local init_globals {
                mata: st_local("init_`g'", st_global("TeXdoc_`g'"))
            }
        }
        _texdoc_do_parse`macval(0)' // returns noclose and cd
        if "`cd'"!="" {
            local pwd `"`c(pwd)'"'
        }
        nobreak {
            capt n break {
                version `caller': texdoc_do`macval(0)'
            }
            if _rc {
                _texdoc_cleanup
                _texdoc_cleanup_do
                if "`cd'"!="" {
                    qui cd `pwd'
                    di as txt `"(cd `pwd')"'
                }
                exit _rc
            }
            if `nested' {
                if `"${TeXdoc_docname}"'=="" {
                    // docname closed (or not yet initialized): skip texdoc 
                    // close and restore previous settings
                    global TeXdoc_stcounter `"`init_stcounter'"'
                    foreach g of local init_globals {
                        global TeXdoc_`g' `"`macval(init_`g')'"'
                    }
                }
                else if `"`init_docname'"'==`"${TeXdoc_docname}"' {
                    // still same docname: skip texdoc close and restore
                    // previous settings, but keep stcounter 
                    foreach g of local init_globals {
                        global TeXdoc_`g' `"`macval(init_`g')'"'
                    }
                }
                else if `"${TeXdoc_docname}"'!="" {
                    // docname has been (re)initialized: 
                    // - if noclose: skip texdoc close and keep settings
                    // - else: apply texdoc close and reset all settings
                    if `"`close'"'=="" {
                        texdoc_close
                        global TeXdoc_stcounter `"`init_stcounter'"'
                        foreach g of local init_globals {
                            global TeXdoc_`g' `"`macval(init_`g')'"'
                        }
                    }
                }
                // reset texdoc do globals
                global TeXdoc_dofile `"`do_dofile'"'
                foreach g of local do_globals {
                    global TeXdoc_do_`g' `"`macval(do_`g')'"'
                }
            }
            else {
                if `"${TeXdoc_docname}"'!="" {
                    if `"`close'"'=="" texdoc_close
                }
                _texdoc_cleanup_do
            }
            if "`cd'"!="" {
                qui cd `pwd'
                di as txt `"(cd `pwd')"'
            }
        }
        exit
    }
    di as err `"`subcmd' invalid subcommand"'
    exit 198
end

program _texdoc_do_parse
    _parse comma fn 0 : 0
    syntax [, noclose cd * ]
    c_local close `close'
    c_local cd `cd'
end

program _texdoc // usage: _texdoc write ...
    mata: texdoc_write_remove_prefix("0")
    tempname fh
    file open `fh' using `"${TeXdoc_docname}"', write append
    file write `fh' `"`macval(0)'"' _n
    file close `fh'
end

program _texdoc_cleanup
    // clear texdoc stlog globals
    foreach g in nodo nolog cmdlog verbatim hardcode nokeep custom cmdstrip ///
        lbstrip noltrim indent name {
        global TeXdoc_st`g' ""
    }
    // clear texdoc init globals
    foreach g in nodo nolog cmdlog verbatim hardcode nokeep custom cmdstrip ///
        lbstrip noltrim gropts grdir stcounter stpath prefix logdir path0 ///
        path basename docname0 docname {
        global TeXdoc_`g' ""
    }
end

program _texdoc_cleanup_do
    // clear texdoc do globals
    foreach g in cmdlogbuf replace append logdir logdir2 noprefix prefix prefix2 ///
        stpath stpath2 nodo nolog cmdlog verbatim hardcode nokeep custom ///
        cmdstrip lbstrip noltrim gropts grdir {
        global TeXdoc_do_`g' ""
    }
    global TeXdoc_dofile ""
end

program texdoc_init
    syntax [anything(id="document name")] [, ///
        Replace Append NOLOGDIR logdir LOGDIR2(str) ///
        NOPrefix Prefix Prefix2(str) NOSTPATH stpath STPATH2(str) ///
        NODO DO NOLOG LOG NOCMDLog CMDLog NOVERBatim VERBatim ///
        NOHardcode Hardcode NOKeep Keep NOCustom Custom NOCMDStrip CMDStrip ///
        NOLBStrip LBStrip NOLTRIM LTRIM GRopts(str asis) grdir(str) ]
    if "`replace'"!="" & "`append'"!="" {
        di as err "replace and append not both allowed"
        exit 198
    }
    foreach opt in logdir prefix stpath {
        if "``opt''``opt'2'"!="" & "`no`opt''"!="" {
            di as err "`opt'() and no`opt' not both allowed"
            exit 198
        }
    }
    foreach opt in do log cmdlog verbatim hardcode keep custom cmdstrip ///
        lbstrip ltrim {
        if "``opt''"!="" & "`no`opt''"!="" {
            di as err "`opt' and no`opt' not both allowed"
            exit 198
        }
    }
    if `"`anything'"'!="" {
        gettoken anything rest : anything // get rid of quotes around filename
        if `"`rest'"'!="" error 198
        // check whether sjlatex is installed
        capt which sjlatex
        if _rc {
            di as err "-sjlatex- package is required. To install the package, type"
            di _n `"    {stata "net install sjlatex, from(http://www.stata-journal.com/production)"}"'
            di ""
            exit 499
        }
        // read global defaults if -texdoc do- is running
        if `"${TeXdoc_dofile}"'!=""{
            if `"`replace'`append'"'=="" {
                local replace `"${TeXdoc_do_replace}"'
                local append `"${TeXdoc_do_append}"'
            }
            if `"`nologdir'`logdir'`logdir2'"'=="" {
                local logdir `"${TeXdoc_do_logdir}"'
                local logdir2 `"${TeXdoc_do_logdir2}"'
            }
            if `"`noprefix'`prefix'`prefix2'"'=="" {
                local noprefix `"${TeXdoc_do_noprefix}"'
                local prefix `"${TeXdoc_do_prefix}"'
                local prefix2 `"${TeXdoc_do_prefix2}"'
            }
            if `"`nostpath'`stpath'`stpath2'"'=="" {
                local stpath `"${TeXdoc_do_stpath}"'
                local stpath2 `"${TeXdoc_do_stpath2}"'
            }
            foreach opt in do log keep ltrim {
                if "``opt''`no`opt''"=="" local no`opt' `"${TeXdoc_do_no`opt'}"'
            }
            foreach opt in cmdlog verbatim hardcode custom cmdstrip lbstrip {
                if "``opt''`no`opt''"=="" local `opt' `"${TeXdoc_do_`opt'}"'
            }
            if `"`macval(gropts)'"'=="" {
                mata: st_local("gropts", st_global("TeXdoc_do_gropts"))
            }
            if `"`grdir'"'==""  local grdir  `"${TeXdoc_do_grdir}"'
        }
        // initialize globals
        mata: texdoc_init()
        // initialize texdoc output file
        tempname fh
        qui file open `fh' using `"${TeXdoc_docname}"', write `replace' `append'
        file close `fh'
        di as txt `"(texdoc output file is ${TeXdoc_docname0})"'
    }
    else if `"${TeXdoc_docname}"'!="" {
        // update globals
        if "`logdir'`logdir2'`nologdir'"!="" {
            if (`"`logdir2'"'!="")      global TeXdoc_logdir `"`logdir2'"'
            else if ("`logdir'"!="") {
                mata: st_global("TeXdoc_logdir", ///
                    pathrmsuffix(st_global("TeXdoc_basename")))
            }
            else                        global TeXdoc_logdir ""
            if `"${TeXdoc_logdir}"'!="" {
                mata: texdoc_mkdir(pathjoin(st_global("TeXdoc_path"), ///
                    st_global("TeXdoc_logdir")))
            }
        }
        if `"`prefix'`prefix2'`noprefix'"'!="" {
            if ("`noprefix'"!="")       global TeXdoc_prefix ""
            else if (`"`prefix2'"'!="") global TeXdoc_prefix `"`prefix2'"'
            else {
                mata: st_global("TeXdoc_prefix", ///
                    pathrmsuffix(st_global("TeXdoc_basename")) + "_")
            }
        }
        if "`stpath'`stpath2'`nostpath'"!="" {
            if (`"`stpath2'"'!="")      global TeXdoc_stpath `"`stpath2'"'
            else if ("`stpath'"!="")    global TeXdoc_stpath `"${TeXdoc_path0}"'
            else                        global TeXdoc_stpath ""
        }
        foreach opt in do log keep ltrim {
            if "``opt''`no`opt''"!="" global TeXdoc_no`opt' `no`opt''
        }
        foreach opt in cmdlog verbatim hardcode custom cmdstrip lbstrip {
            if "``opt''`no`opt''"!="" global TeXdoc_`opt' ``opt''
        }
        if `"`gropts'"'!="" {
            global TeXdoc_gropts `"`macval(gropts)'"'
        }
        if `"`grdir'"'!=""  global TeXdoc_grdir `"`grdir'"'
        if `"${TeXdoc_grdir}"'!="" {
            mata: texdoc_mkdir(pathjoin(st_global("TeXdoc_path"), ///
                st_global("TeXdoc_grdir")))
        }
    }
    else {
        di as txt "(texdoc not initialized; nothing to do)"
        exit
    }
    // clear s-returns (will be filled by texdoc close)
    sreturn clear
end

program texdoc_close, sclass
    if `"`macval(0)'"'!="" error 198
    if `"${TeXdoc_docname}"'!="" {
        di as txt `"(texdoc output written to {browse `"${TeXdoc_docname}"':${TeXdoc_docname0}})"'
        sreturn clear
        foreach s in custom nokeep hardcode verbatim cmdlog noltrim lbstrip ///
            cmdstrip nolog nodo gropts grdir stpath prefix logdir path      ///
            basename docname {
            mata: st_local("`s'", st_global("TeXdoc_`s'"))
            sreturn local `s' `"`macval(`s')'"'
        }
    }
    else {
        di as txt "(texdoc not initialized; nothing to do)"
        exit
    }
    _texdoc_cleanup
end

program texdoc_append
    local 0 `"using `macval(0)'"'
    syntax using/
    if `"${TeXdoc_docname}"'=="" {
        di as txt "(texdoc not initialized; nothing to do)"
        exit
    }
    tempname fh1 fh2
    file open `fh1' using `"${TeXdoc_docname}"', write append
    file open `fh2' using `"`using'"', read
    file read `fh2' line
    while r(eof)==0 {
        file write `fh1' `"`macval(line)'"' _n
        file read `fh2' line
    }
    file close `fh1'
    file close `fh2'
    di as txt `"(`using' appended)"'
end

program texdoc_stlog
    version 9.2
    local caller : di _caller()
    gettoken subcmd args : 0, parse(", ")
    local length = length(`"`subcmd'"')
    if `"`subcmd'"'==substr("close",1,max(`length',1)) {
        version `caller': texdoc_stlog_close`macval(args)'
    }
    else if `"`subcmd'"'==substr("oom",1,max(`length',1)) {
        version `caller': texdoc_stlog_oom`macval(args)'
    }
    else if `"`subcmd'"'=="cnp" {
        texdoc_stlog_cnp`macval(args)'
    }
    else {
        version `caller': texdoc_stlog_open `macval(0)'
    }
end

program texdoc_stlog_open
    version 9.2
    local caller : di _caller()
    if `"${TeXdoc_docname}"'=="" {
        di as txt "(texdoc not initialized; nothing to do)"
        exit
    }
    syntax [anything(name=stname)] [using/] [, ///
        NODO DO NOLOG LOG NOCMDLog CMDLog NOVERBatim VERBatim ///
        NOHardcode Hardcode NOKeep Keep NOCustom Custom NOCMDStrip CMDStrip ///
        NOLBStrip LBStrip NOLTRIM LTRIM ]
    foreach opt in do log keep ltrim {
        if "``opt''"!="" & "`no`opt''"!="" {
            di as err "`opt' and no`opt' not both allowed"
            exit 198
        }
        if "``opt''`no`opt''"=="" local no`opt' ${TeXdoc_no`opt'}
        global TeXdoc_stno`opt' `no`opt''
    }
    foreach opt in cmdlog verbatim hardcode custom cmdstrip lbstrip {
        if "``opt''"!="" & "`no`opt''"!="" {
            di as err "`opt' and no`opt' not both allowed"
            exit 198
        }
        if "``opt''`no`opt''"=="" local `opt' ${TeXdoc_`opt'}
        global TeXdoc_st`opt' ``opt''
    }
    if `"`custom'"'!="" {  // custom implies nohardcode
        local hardcode
        global TeXdoc_sthardcode ""
    }
    if `"`stname'"'!="" {
        global TeXdoc_stname `"`stname'"'
    }
    else {
        local cnterr 0
        if `"${TeXdoc_stcounter}"'=="" {
            local cnterr 1
        }
        else {
            capt confirm integer number ${TeXdoc_stcounter}
            if _rc local cnterr 1
            else {
                if ${TeXdoc_stcounter} < 0 local cnterr 1
            }
        }
        if `cnterr' {
            di as error "something's wrong; invalid texdoc stlog counter"
            exit 499
        }
        global TeXdoc_stcounter = ${TeXdoc_stcounter} + 1
        global TeXdoc_stname    `"${TeXdoc_prefix}${TeXdoc_stcounter}"'
    }
    mata: st_local("tmp", pathjoin(st_global("TeXdoc_logdir"), st_global("TeXdoc_stname")))
    if "`nolog'`nodo'`cmdlog'"=="" {
        di as txt `"(opening texdoc stlog `tmp')"'
        mata: st_local("tmp", ///
            pathjoin(st_global("TeXdoc_path"), st_local("tmp")))
        version `caller': sjlog using `"`tmp'"', replace
    }
    else if "`nolog'"=="" & "`cmdlog'"!="" {
        mata: st_local("tmp", ///
            pathjoin(st_global("TeXdoc_path"), st_local("tmp")))
        if "`verbatim'"!="" local tmp `"`tmp'.log.tex"'
        if `"`using'"'!="" {
            qui copy `"`using'"' `"`tmp'"', replace
        }
        else {
            capt copy `"${TeXdoc_do_cmdlogbuf}"' `"`tmp'"', replace
        }
    }
    if `"`using'"'!="" {
        if "`nodo'"=="" {
            do `"`using'"'
        }
        version `caller': texdoc_stlog_close
    }
end

program texdoc_stlog_oom
    version 9.2
    local caller : di _caller()
    version `caller': quietly `macval(0)'
    di _n as txt "\TeXdoc_OOM"
end

program texdoc_stlog_cnp
    if `"`macval(0)'"'!="" error 198
    di as txt "\TeXdoc_CNP"
end

program texdoc_stlog_close, sclass
    version 9.2
    local caller : di _caller()
    if `"`macval(0)'"'!="" error 198
    if `"${TeXdoc_docname}"'=="" {
        di as txt "(texdoc not initialized; nothing to do)"
        exit
    }
    foreach opt in nodo nolog cmdlog verbatim hardcode nokeep custom ///
        cmdstrip lbstrip noltrim indent {
        local `opt' ${TeXdoc_st`opt'}
        global TeXdoc_st`opt' ""
    }
    local stname0 `"${TeXdoc_stname}"'
    global TeXdoc_stname ""
    mata: st_local("stname", pathjoin(st_global("TeXdoc_logdir"), st_local("stname0")))
    mata: st_local("logname0", pathjoin(st_global("TeXdoc_path"), st_local("stname")))
    mata: st_local("logpath0", pathjoin(st_global("TeXdoc_path0"), st_local("stname")))
    mata: st_local("texname0", pathjoin(st_global("TeXdoc_stpath"), st_local("stname")))
    local logname `"`logname0'.log.tex"'
    local logpath0 `"`logpath0'.log.tex"'
    local texname `"`texname0'.log.tex"'
    if "`nolog'`nodo'`cmdlog'"=="" {
        version `caller': sjlog close, replace noclean nologfile
    }
    if `"`stname'"'!="" & "`nolog'"=="" {
        // process log file and cleanup
        if "`cmdlog'"!="" {
            if "`verbatim'"=="" {
                capt confirm file `"`logname0'"'
                if _rc==0 {
                    capt confirm file `"`logname'"'
                    if _rc local sjlogreplace
                    else   local sjlogreplace ", replace"
                    version `caller': sjlog type `"`logname0'"'`sjlogreplace'
                    erase `"`logname0'"'
                }
                else {
                    capt erase `"`logname'"'
                }
            }
        }
        else if "`nodo'"=="" {
            capt erase `"`logname0'.smcl"'
            nobreak {
                capt n break {
                    mata: texdoc_instance_fh("fh")
                    mata: texdoc_striplog(`fh'=.)
                }
                local rc = _rc 
                capture mata: texdoc_closeout_fh(`fh')
                capture mata: mata drop `fh'
                if `rc' exit `rc'
            }
        }
        // TeX insert
        if `"`indent'"'!="" {
            mata: st_local("indent", " " * `indent')
            // local indent will be returned in s(), but no indentation will
            // be applied to log files as it can cause spacing issues
        } 
        if "`custom'"!="" {
            if "`nodo'"=="" | "`cmdlog'"!="" {
                di as txt `"(texdoc stlog written to {browse `"`logname'"':`logpath0'})"'
            }
        }
        else if "`hardcode'"!="" {
            if "`nokeep'"=="" & ("`nodo'"=="" | "`cmdlog'"!="") {
                di as txt `"(texdoc stlog written to {browse `"`logname'"':`logpath0'})"'
            }
            if "`cmdlog'"!="" & "`verbatim'"!="" {
                _texdoc write \begin{stverbatim}
                _texdoc write \begin{verbatim}
            } 
            else _texdoc write \begin{stlog}
            capt confirm file `"`logname'"'
            if _rc {
                di as txt `"(`logname' not found)"'
            }
            else {
                quietly texdoc_append `"`logname'"'
                di as txt `"(texdoc stlog `stname' appended)"'
            }
            if "`cmdlog'"!="" & "`verbatim'"!="" {
                _texdoc write \end{verbatim}
                _texdoc write \end{stverbatim}
            } 
            else _texdoc write \end{stlog}
            if "`nokeep'"!="" {
                capt erase `"`logname'"'
            }
        }
        else {
            if "`nodo'"=="" | "`cmdlog'"!="" {
                di as txt `"(texdoc stlog written to {browse `"`logname'"':`logpath0'})"'
            }
            if "`cmdlog'"!="" & "`verbatim'"!="" {
                _texdoc write \begin{stverbatim}
                _texdoc write \verbatiminput{`texname'}
                _texdoc write \end{stverbatim}
            }
            else {
                if strpos(`"`texname'"', "-") { // names with "-" problem for \input{}
                    _texdoc write {\def\TeXdocstname{`texname'}
                    _texdoc write \begin{stlog}
                    _texdoc write \input{\TeXdocstname}
                    _texdoc write \end{stlog}}
                }
                else {
                    _texdoc write \begin{stlog}
                    _texdoc write \input{`texname'}
                    _texdoc write \end{stlog}
                }
            }
        }
    }
    sreturn clear
    foreach s in custom nokeep hardcode verbatim cmdlog noltrim lbstrip ///
        cmdstrip nolog nodo indent texname0 texname {
        sreturn local `s' `"``s''"'
    }
    sreturn local filename0 `"`logname0'"'
    sreturn local filename  `"`logname'"'
    sreturn local name0     `"`stname0'"'
    sreturn local name      `"`stname'"'
end

program texdoc_graph
    // anything to do?
    if `"${TeXdoc_docname}"'=="" {
        di as txt "(texdoc not initialized; nothing to do)"
        exit
    }
    if `"`s(name)'"'=="" | `"`s(filename0)'"'=="" | `"`s(texname0)'"'=="" {
        di as txt "(no results from texdoc stlog available; nothing to do)"
        exit
    }
    
    // copy s() info
    local export = (`"`s(nodo)'"'=="")
    local stname    `"`s(name)'"'
    local stname0   `"`s(name0)'"'
    local filename0 `"`s(filename0)'"'
    local texname0  `"`s(texname0)'"'
    local indent    `"`s(indent)'"'
    
    // syntax
    syntax [, * ]
    _texdoc_graph_syntax, `macval(options)'
    // get default options
    mata: st_local("opt_options", st_global("TeXdoc_gropts"))
    _texdoc_graph_syntax opt_, `macval(opt_options)'
    foreach opt in as optargs name {
        if `"`macval(`opt')'"'=="" {
            local `opt' `macval(opt_`opt')'
        }
    }
    if "`figure'`nofigure'"=="" {
        local nofigure `opt_nofigure'
        local figure   `opt_figure'
        local figure2  `macval(opt_figure2)'
    }
    if "`figure'"!="" & `"`macval(figure2)'"'=="" {
        local figure2 `macval(opt_figure2)'
    }
    foreach opt in epsfig suffix center custom {
        if `"``opt''`no`opt''"'=="" {
            local `opt'   `opt_`opt''
            local no`opt' `opt_no`opt''
        }
    }
    foreach opt in caption label {
        if `"`macval(`opt')'"'=="" {
            local `opt'   `macval(opt_`opt')'
        }
    }
    if "`cabove'`cbelow'"=="" {
        local cabove `opt_cabove'
        local cbelow `opt_cbelow'
    }
    if `"`options'"'=="" local options `opt_options'
    // strip outer quotes from options with arguments
    _texdoc_graph_syntax2, `as' `macval(optargs)' `macval(figure2)' ///
        `macval(caption)' `macval(label)' `name'
    // set defaults
    if "`nocenter'"==""         local center center    // center is default
    if "`cabove'`cbelow'"==""   local cbelow cbelow    // cbelow is default
    if `"`macval(caption)'`macval(label)'"'!="" & "`nofigure'"=="" {
        local figure figure                 // center and label imply figure
    }
    if `"`as'"'=="" {
        if "`epsfig'"!="" local as "eps"
        else              local as "pdf"
    }
    if "`nosuffix'"=="" {
        if `:list sizeof as'==1 {
            local suffix `".`as'"'
        }
        else {
            if "`suffix'"!="" {
                local suffix: word 1 of `as'
                local suffix `".`suffix'"'
            }
        }
    }
    else local suffix
    if `"`name'"'!="" local name name(`name')
    
    // export graph
    if `export' {
        foreach ff of local as {
            if `"${TeXdoc_grdir}"'!="" {
                local grname `"`stname0'.`ff'"'
                mata: st_local("grname", pathjoin(st_global("TeXdoc_grdir"), st_local("grname")))
                mata: st_local("filename", pathjoin(st_global("TeXdoc_path"), st_local("grname")))
            }
            else {
                local grname   `"`stname'.`ff'"'
                local filename `"`filename0'.`ff'"'
            }
        mata: st_local("grname", pathjoin(st_global("TeXdoc_path0"), st_local("grname")))
        quietly graph export `"`filename'"', replace `name' `options'
        di as txt `"(graph written to {browse `"`filename'"':`grname'})"'
        }
    }
    
    // include link in document
    if "`custom'"!="" exit
    if `"${TeXdoc_grdir}"'!="" {
        local texname `"`stname0'`suffix'"'
        mata: st_local("texname", pathjoin(st_global("TeXdoc_grdir"), ///
            st_local("texname")))
        mata: st_local("texname", pathjoin(st_global("TeXdoc_stpath"), ///
            st_local("texname")))
    }
    else local texname  `"`texname0'`suffix'"'
    local indent1
    if "`figure'`center'"!=""  _texdoc write     // add an empty line
    if "`figure'"!="" {
        if `"`figure2'"'!="" {
                               _texdoc write `indent'\begin{figure}[`macval(figure2)']
        }  
        else                   _texdoc write `indent'\begin{figure}
        local indent1 "    "
        if "`center'"!=""      _texdoc write `indent'`indent1'\centering
    }
    else if "`center'"!="" {
                               _texdoc write `indent'\begin{center}
        local indent1 "    "
    }
    if "`cabove'"!="" {
        if `"`caption'"'!="" {
                               _texdoc write `indent'`indent1'\caption{`macval(caption)'}
        }
        if `"`label'"'!=""     {
                               _texdoc write `indent'`indent1'\label{`macval(label)'}
        }
    }
    if "`epsfig'"!="" {
        if `"`optargs'"'!="" {
                               _texdoc write `indent'`indent1'\epsfig{file=`texname',`macval(optargs)'}
        }
        else                   _texdoc write `indent'`indent1'\epsfig{file=`texname'}
    }
    else {
        if `"`optargs'"'!="" {
                               _texdoc write `indent'`indent1'\includegraphics[`macval(optargs)']{`texname'}
        }
        else                   _texdoc write `indent'`indent1'\includegraphics{`texname'}
    }
    if "`cbelow'"!="" {
        if `"`caption'"'!="" {
                               _texdoc write `indent'`indent1'\caption{`macval(caption)'}
        }
        if `"`label'"'!=""     {
                               _texdoc write `indent'`indent1'\label{`macval(label)'}
        }
    }
    if "`figure'"=="" {
        if "`center'"!=""      _texdoc write `indent'\end{center}
    }
    else                       _texdoc write `indent'\end{figure}
end
program _texdoc_graph_syntax
    syntax [anything(name=prefix)] [,                           ///
        as(passthru) Optargs(passthru) NOSuffix Suffix          ///
        NOEPSFIG epsfig NOCenter Center                         ///
        NOFigure Figure Figure2(passthru) NOCUSTOM custom       ///
        CAPtion(passthru) Label(passthru) CAbove CBelow         ///
        name(passthru) * ]
    if "`cabove'"!="" & "`cbelow'"!="" {
        di as err "cabove and cbelow not both allowed"
        exit 198
    }
    if `"`macval(figure2)'"'!="" local figure figure
    foreach opt in suffix epsfig center figure custom {
        if "``opt''"!="" & "`no`opt''"!="" {
            di as err "`opt' and no`opt' not both allowed"
            exit 198
        }
    }
    foreach opt in as optargs nosuffix suffix noepsfig epsfig   ///
        nocenter center nofigure figure figure2 nocustom custom ///
        caption label cabove cbelow name options {
        c_local `prefix'`opt' `"`macval(`opt')'"'
    }
end
program _texdoc_graph_syntax2
    syntax [, as(str) optargs(str) figure2(str) caption(str) ///
        label(str) name(str) ]
    foreach opt in as optargs figure2 caption label name {
        c_local `opt' `"`macval(`opt')'"'
    }
end

program texdoc_strip
    syntax [anything] [, Replace Append ]
    if "`replace'"!="" & "`append'"!="" {
        di as err "replace and append not both allowed"
        exit 198
    }
    gettoken in out : anything
    gettoken out rest : out
    if `"`rest'"'!="" error 198
    if `"`in'"'=="" | `"`out'"'==""  {
        di as err "must specify input filename and output filename"
        exit 198
    }
    if `"`replace'`append'"'=="" confirm new file `"`out'"'
    nobreak {
        capt n break {
            mata: texdoc_instance_fh("fh")
            mata: texdoc_strip(`fh'=.)
        }
        local rc = _rc 
        capture mata: texdoc_closeout_fh(`fh')
        capture mata: mata drop `fh'
        if `rc' exit `rc'
    }
    di as txt `"(output written to {browse `"`out'"'})"'
end

program texdoc_do
    version 9.2
    local caller : di _caller()
    syntax anything(name=dofile id="filename") ///
        [, Init(str) cd noclose Savecmd(str asis) * ]
    _texdoc_do, `macval(options)' // set global defaults
    if `"${TeXdoc_dofile}"'!="" & `"`init'"'=="" & `"${TeXdoc_docname}"'!="" {
        // update globals if nested and document already open
        texdoc_init, `macval(options)'
    }
    gettoken dofile rest : dofile // get rid of quotes around filename
    if `"`rest'"'!="" error 198
    mata: texdoc_add_suffix("dofile", ".do")
    mata: texdoc_add_abspath("dofile")
    global TeXdoc_dofile `"`dofile'"'
    if "`cd'"!="" {
        mata: texdoc_get_path(st_local("dofile")) // returns local path
        qui cd `"`path'"'
        di as txt `"(cd `path')"'
    }
    if `"`init'"'!="" {
        texdoc_init `"`init'"'
    }
    tempfile dobuf
    tempfile cmdlogbuf // for cmdlog option
    local hascmdlogbuf 0
    local flushdobuf 0
    local offset 0
    local texmode 0
    if `"`savecmd'"'!="" {
        _texdoc_parse_savecmd `savecmd'
        mata: texdoc_add_suffix("savecmd", ".do")
        local savecmd0: copy local savecmd
        mata: texdoc_add_abspath("savecmd")
        if `"`dofile'"'==`"`savecmd'"' {
            di as err "input file and savecmd() file must be different"
            exit 498
        }
        tempname fh
        qui file open `fh' using `"`savecmd'"', write `screplace' `scappend'
        file close `fh'
    }
    nobreak {
        capt n break {
            global TeXdoc_do_cmdlogbuf `"`cmdlogbuf'"'
            mata: texdoc_instance_fh("fh")
            while (`offset'<.) {
                mata: texdoc_do(`fh'=.)
                if (`flushdobuf') {
                    version `caller': do `"`dobuf'"'
                    local flushdobuf 0
                }
            }
        }
        local rc = _rc
        capture mata: texdoc_closeout_fh(`fh')
        capture mata: mata drop `fh'
        global TeXdoc_do_cmdlogbuf ""
        if `rc' exit `rc'
    }
    if `"`savecmd'"'!="" {
        di as txt `"(Stata commands written to {browse `"`savecmd'"':`savecmd0'})"'
    }
end

program _texdoc_do
    syntax [, Replace Append NOLOGDIR logdir LOGDIR2(str) ///
        NOPrefix Prefix Prefix2(str) NOSTPATH stpath STPATH2(str) ///
        NODO DO NOLOG LOG NOCMDLog CMDLog NOVERBatim VERBatim ///
        NOHardcode Hardcode NOKeep Keep NOCustom Custom NOCMDStrip CMDStrip ///
        NOLBStrip LBStrip NOLTRIM LTRIM GRopts(str asis) grdir(str) ]
    if "`replace'"!="" & "`append'"!="" {
        di as err "replace and append not both allowed"
        exit 198
    }
    foreach opt in logdir prefix stpath {
        if "``opt''``opt'2'"!="" & "`no`opt''"!="" {
            di as err "`opt'() and no`opt' not both allowed"
            exit 198
        }
    }
    foreach opt in do log cmdlog verbatim hardcode keep custom cmdstrip ///
        lbstrip ltrim {
        if "``opt''"!="" & "`no`opt''"!="" {
            di as err "`opt' and no`opt' not both allowed"
            exit 198
        }
    }
    // read global defaults if nested
    if `"${TeXdoc_dofile}"'!=""{
        if `"`replace'`append'"'=="" {
            local replace `"${TeXdoc_do_replace}"'
            local append `"${TeXdoc_do_append}"'
        }
        if `"`nologdir'`logdir'`logdir2'"'=="" {
            local logdir `"${TeXdoc_do_logdir}"'
            local logdir2 `"${TeXdoc_do_logdir2}"'
        }
        if `"`noprefix'`prefix'`prefix2'"'=="" {
            local noprefix `"${TeXdoc_do_noprefix}"'
            local prefix `"${TeXdoc_do_prefix}"'
            local prefix2 `"${TeXdoc_do_prefix2}"'
        }
        if `"`nostpath'`stpath'`stpath2'"'=="" {
            local stpath `"${TeXdoc_do_stpath}"'
            local stpath2 `"${TeXdoc_do_stpath2}"'
        }
        foreach opt in do log keep ltrim {
            if "``opt''`no`opt''"=="" local no`opt' `"${TeXdoc_do_no`opt'}"'
        }
        foreach opt in cmdlog verbatim hardcode custom cmdstrip lbstrip {
            if "``opt''`no`opt''"=="" local `opt' `"${TeXdoc_do_`opt'}"'
        }
        if `"`macval(gropts)'"'=="" {
            mata: st_local("gropts", st_global("TeXdoc_do_gropts"))
        }
        if `"`grdir'"'==""  local grdir `"${TeXdoc_do_grdir}"'
    }
    foreach opt in replace append logdir logdir2 noprefix prefix prefix2 stpath ///
        stpath2 nodo nolog cmdlog verbatim hardcode nokeep custom cmdstrip ///
        lbstrip noltrim gropts grdir {
        global TeXdoc_do_`opt' `"`macval(`opt')'"'
    }
end

program _texdoc_parse_savecmd
    syntax anything(id="filename") [, Replace Append ]
    if "`replace'"!="" & "`append'"!="" {
        di as err "savecmd(): replace and append not both allowed"
        exit 198
    }
    c_local savecmd `"`anything'"'
    c_local screplace `replace'
    c_local scappend `append'
end

version 9.2
mata:
mata set matastrict on

void texdoc_write_remove_prefix(string scalar l)
{   // removes leading "write" and one blank
    string scalar s
    
    s = st_local(l)
    if (substr(s, 6, 1)==" ") st_local(l, substr(s, 7, .))
    else                      st_local(l, substr(s, 6, .))
}

void texdoc_init()
{
    string scalar fn, path, path0, fname
    
    fn = st_local("anything")
    pathsplit(fn, path, fname)
    path0 = path
    if (pathisabs(path)) texdoc_mkdir(path)
    else {
        texdoc_mkdir(path)
        path = pathjoin(pwd(), path)
    }
    if (pathsuffix(fname)=="") fname = fname + ".tex"
    fn = pathjoin(path, fname)
    if (fn==st_global("TeXdoc_dofile")) {
        display("{err}init docname must be different from source do-file")
        exit(error(498))
    }
    st_global("TeXdoc_docname",       fn)       // filename with abs. path
    st_global("TeXdoc_docname0",      pathjoin(path0, fname))
    st_global("TeXdoc_basename",      fname)
    st_global("TeXdoc_path",          path)     // absolute path
    st_global("TeXdoc_path0",         path0)    // path as specified
    if (st_local("logdir2")!="")      st_global("TeXdoc_logdir", st_local("logdir2"))
    else if (st_local("logdir")!="")  st_global("TeXdoc_logdir", pathrmsuffix(fname))
    else                              st_global("TeXdoc_logdir", "")
    if (st_global("TeXdoc_logdir")!="") {
        texdoc_mkdir(pathjoin(path0, st_global("TeXdoc_logdir")))
    }
    if (st_local("noprefix")!="")     st_global("TeXdoc_prefix", "")
    else if (st_local("prefix2")!="") st_global("TeXdoc_prefix", st_local("prefix2"))
    else if (st_local("prefix")!="")  st_global("TeXdoc_prefix", pathrmsuffix(fname) + "_")
    else if (st_global("TeXdoc_logdir")!="") st_global("TeXdoc_prefix", "")
    else                              st_global("TeXdoc_prefix", pathrmsuffix(fname) + "_")
    if (st_local("stpath2")!="")      st_global("TeXdoc_stpath", st_local("stpath2"))
    else if (st_local("stpath")!="")  st_global("TeXdoc_stpath", path0)
    else                              st_global("TeXdoc_stpath", "")
    st_global("TeXdoc_stcounter",     "0")
    st_global("TeXdoc_stname",        "")
    st_global("TeXdoc_nodo",          st_local("nodo"))
    st_global("TeXdoc_nolog",         st_local("nolog"))
    st_global("TeXdoc_cmdlog",        st_local("cmdlog"))
    st_global("TeXdoc_verbatim",      st_local("verbatim"))
    st_global("TeXdoc_hardcode",      st_local("hardcode"))
    st_global("TeXdoc_nokeep",        st_local("nokeep"))
    st_global("TeXdoc_custom",        st_local("custom"))
    st_global("TeXdoc_cmdstrip",      st_local("cmdstrip"))
    st_global("TeXdoc_lbstrip",       st_local("lbstrip"))
    st_global("TeXdoc_noltrim",       st_local("noltrim"))
    st_global("TeXdoc_gropts",        st_local("gropts"))
    st_global("TeXdoc_grdir",         st_local("grdir"))
    if (st_global("TeXdoc_grdir")!="") {
        texdoc_mkdir(pathjoin(path0, st_global("TeXdoc_grdir")))
    }
}

void texdoc_mkdir(string scalar path)
{
    if (direxists(path)==0) {
        mkdir(path)
        printf("{txt}(directory '%s' created)\n", path)
    }
}

void texdoc_striplog(real scalar fh)
{
    real scalar      i, r, a, lbstrip, cmdstrip, lb, cb, lskip
    real colvector   p
    string scalar    fn, s
    string rowvector f

    lbstrip = (st_local("lbstrip")!="")
    cmdstrip = (st_local("cmdstrip")!="")
    fn = st_local("logname")
    f = texdoc_cat(fn, fh)
    r = rows(f)
    if (r<1) return
    p = J(r,1,1)
    if (f[1]=="{\smallskip}") p[1] = 0  // strip first line
    for (i=1; i<=r; i++) {
        s = f[i]
        if      (s=="\\TeXdoc_OOM") f[i] = "\oom"
        else if (s=="\\TeXdoc_CNP") f[i] = "\cnp"
        else if (substr(s,1,2)==". ") {
            a = i
            lb = cb = 0
            s = texdoc_strip_comments(substr(s,3,.), lb, cb)
            while (1) {
                if (i==r) break
                if (substr(f[i+1],1,2)!="> ") break
                i++
                s = s + texdoc_strip_comments(substr(f[i],3,.), lb, cb)
            }
            if (cmdstrip) {
                p[|a \ i|] = J(i-a+1, 1, 0)
                continue
            }
            lskip = strlen(s)
            s = strltrim(s)
            lskip = lskip - strlen(s)
            if (strrtrim(substr(s, 1, 7))=="texdoc") {  // texdoc ...
                s = tokens(substr(s, 8, .))
                if (length(s)==2) {
                    if (i==r) {
                        if (texdoc_cmdmatch(s, ("stlog", "close"), (1, 1))) {
                            p[|a \ .|] = J(r-a+1, 1, 0)
                            break   // end of logfile
                        }
                    }
                    if (texdoc_cmdmatch(s, ("stlog", "cnp"), (1, 3))) {
                        p[|a \ i|] = J(i-a+1, 1, 0)
                        continue
                    }
                }
                else if (length(s)>=3) {
                    if (texdoc_cmdmatch(s, ("stlog", "oom"), (1, 1))) {
                        texdoc_strip_oom(f, p, a, i, lskip, 1)
                    }
                }
                else i = a
            }
            else i = a
        }
        if (lbstrip) {
            if (substr(f[i],1,2)==". ") {
                lb = cb = 0
                s = texdoc_strip_comments(f[i], lb, cb)
                while (1) {
                    if (i==r) break
                    if (substr(f[i+1],1,2)!="> ") break
                    if (cb==0) {
                        lb = texdoc_striplog_lbpos(f[i], cb)-1
                        if (lb>0) f[i] = substr(f[i], 1, lb)
                    }
                    i++
                    s = texdoc_strip_comments(f[i], lb, cb)
                }
            }
        }
    }
    if (r>1) {
        if (f[|r-1 \ .|]==(". ", "end of do-file")') {
            p[|r-1 \ .|] = J(2, 1, 0)    // if -texdoc stlog using-
        }
    }
    unlink(fn)
    texdoc_fput(fn, fh, select(f, p), "w")
}

void texdoc_strip(real scalar fh)
{
    real scalar      i, r, texmode, tl, stlog, lb, cb, a, lskip
    real colvector   p
    string scalar    in, out, tex1, tex2, texstop, texstart, texdocexit
    string rowvector s
    string colvector f

    in = st_local("in"); out = st_local("out")
    texdocexit = "// texdoc exit"
    tex1 = "/*tex"
    tex2 = "/***"
    texmode = stlog = lb = cb = 0

    f = texdoc_cat(in, fh)
    r = rows(f)
    if (r<1) return
    p = J(r,1,1)
    for (i=1; i<=r; i++) {
        if (lb==0) {                                // new command line
            s = subinstr(f[i], char(9), " ")        // expand tabs
            if (texmode) {                          // process tex block
                p[i] = 0
                s = strtrim(s)
                if (texmode<.) {
                    if (strrtrim(substr(s,1,tl))==texstart) texmode++
                }
                else texmode = 1
                if ((strlen(s)>=tl ? strltrim(substr(s,-tl,.)) : s)==texstop) texmode--
                continue
            }
            if (stlog==0) {                         // start of tex block
                if (strrtrim(substr(strltrim(s),1,6))==tex1) {
                    texstart = "/*tex"
                    texstop  = "tex*/"
                    tl = 6
                    p[i] = 0
                    texmode = .
                    i-- // closing tag may be on same line
                    continue
                }
                if (strrtrim(substr(strltrim(s),1,5))==tex2) {
                    texstart = "/***"
                    texstop  = "***/"
                    tl = 5
                    p[i] = 0
                    texmode = .
                    i-- // closing tag may be on same line
                    continue
                }
            }
            if (strtrim(stritrim(s))==texdocexit) { // end of input
                p[|i \ .|] = J(r-i+1, 1, 0)
                break
            }
            a = i
            s = texdoc_strip_comments(s, lb, cb)
        }
        else {                                      // continued command line
            s = s + texdoc_strip_comments(subinstr(f[i], char(9), " "), lb, cb)
        }
        if (lb) continue                            // command not complete yet
        lskip = strlen(s) - strlen(strltrim(s))
        s = strtrim(stritrim(s))
        if (strrtrim(substr(s, 1, 4))=="tex") {     // tex ...
            p[|a \ i|] = J(i-a+1, 1, 0)
            continue
        }
        if (strrtrim(substr(s, 1, 7))=="texdoc") {  // texdoc ...
            s = tokens(substr(s, 8, .), ", ")
            if (texdoc_cmdmatch(s, "write", 1)) {
                p[|a \ i|] = J(i-a+1, 1, 0)
            }
            else if (texdoc_cmdmatch(s, "stlog", 1)) {
                if (texdoc_cmdmatch(s, ("stlog", "close"), (1, 1))) {
                    stlog = 0
                    p[|a \ i|] = J(i-a+1, 1, 0)
                }
                else if (texdoc_cmdmatch(s, ("stlog", "oom"), (1, 1))) {
                    texdoc_strip_oom(f, p, a, i, lskip, 0)
                }
                else if (texdoc_cmdmatch(s, ("stlog", "cnp"), (1, 3))) {
                    p[|a \ i|] = J(i-a+1, 1, 0)
                }
                else {
                    stlog = 1
                    p[|a \ i|] = J(i-a+1, 1, 0)
                    _texdoc_strip_using(f, p, a, s)
                    if (p[a]==1) stlog = 0
                }
            }
            else if (texdoc_cmdmatch(s, "do", 2)) {
                p[|a \ i|] = J(i-a+1, 1, 0)
                p[a] = 1; f[a] = "do"
                _texdoc_strip_opts(f, a, s, 1)
            }
            else p[|a \ i|] = J(i-a+1, 1, 0)
        }
    }
    if (st_local("replace")!="") unlink(out)
    if (st_local("append")!="")  texdoc_fput(out, fh, select(f, p), "a")
    else                         texdoc_fput(out, fh, select(f, p), "w")
}

void _texdoc_strip_using(string colvector f, real colvector p, real scalar a, 
    string rowvector s) // parse: -... using fn, ...- and store -do fn-
{
    real scalar j
    
    for (j=2; j<=length(s); j++) {
        if (s[j]=="using") {
            p[a] = 1
            f[a] = "do"
            _texdoc_strip_opts(f, a, s, j)
            break
        }
        if (s[j]==",") break
    }
}

void _texdoc_strip_opts(string colvector f, real scalar a, string rowvector s, 
    real scalar j) // strip ", ..." from "fn, ..." and store quoted fn
{
    for (j++; j<=length(s); j++) {
        if (s[j]==",") break
        if (strpos(s[j], `"""'))
            f[a] = f[a] + " `" + `"""' + s[j] + `"""' + "'"
        else if (strpos(s[j], " "))
            f[a] = f[a] + `" ""' + s[j] + `"""'
        else if (s[j]=="")
            f[a] = f[a] + `" ""' + `"""'
        else f[a] = f[a] + " " + s[j]
    }
}

void texdoc_do(real scalar fh)
{
    real scalar texmode
    
    texmode = strtoreal(st_local("texmode"))
    if (texmode) texdoc_do_texblock(texmode, fh)
    else         texdoc_do_doblock(texmode, fh)
    st_local("texmode", strofreal(texmode))
}

void texdoc_do_texblock(real scalar texmode, real scalar fh)
{
    real scalar      offset, i, l, tl, EOF
    string scalar    s, texdoc, line, texstart, texstop
    string colvector lines
    
    texdoc = st_global("TeXdoc_docname")
    texstart = "/*" + st_local("textag")
    texstop  = st_local("textag") + "*/"
    tl = strlen(texstart) + 1
    EOF = 0
    offset = strtoreal(st_local("offset"))
    fh = fopen(st_local("dofile"), "r")
    fseek(fh, offset, -1)
    
    line = fget(fh)
    s = strtrim(subinstr(line, char(9), " ")) // expand tabs and trim
    l = 1
    if (strltrim(substr(s,-tl,.))==texstop) texmode--
    else {                                  // find end of tex block
        while ((line = fget(fh))!=J(0,0,"")) {
            l++
            s = strtrim(subinstr(line, char(9), " "))
            if (strrtrim(substr(s,1,tl))==texstart) texmode++
            if ((strlen(s)>=tl ? strltrim(substr(s,-tl,.)) : s)==texstop) {
                if (_texdoc_line_has_texcomments(s)==0) texmode--
            }
            if (texmode==0) break
        }
        if (line==J(0,0,"")) EOF = 1
    }
    fseek(fh, offset, -1)
    lines = J(l, 1, "")
    for (i=1; i<=l; i++) {              // read tex lines
        line = fget(fh)
        if (i==1) line = substr(line, strpos(line, texstart) + tl, .)
        if (i==l) line = substr(line, 1, _texdoc_strrpos(line, texstop) - 2)
        lines[i] = line
    }
    if (l>1) {                          // remove first and last line if empty
        if (strtrim(lines[l])=="") lines = lines[|1 \ l-1|]
        if (strtrim(lines[1])=="") {
            if (rows(lines)==1) lines = J(0, 1, "")
            else lines = lines[|2 \ .|]
        }
    }
    if (EOF) st_local("offset", ".")
    else     st_local("offset", strofreal(ftell(fh)))
    fclose(fh)
    if (texdoc!="") texdoc_fput(texdoc, fh, lines, "a")
}
real scalar _texdoc_line_has_texcomments(string scalar s)
{
    real scalar p, p1
    
    p = strpos(s,"%")
    if (p==0) return(0)
    while (1) {
        if (substr(s, p-1, 1)!="\") return(1)
        p1 = strpos(substr(s, p+1, .), "%")
        if (p1==0) return(0)
        p = p + p1
    }
}
real scalar _texdoc_strrpos(string scalar s0, string scalar s)
{   // because strrpos() not available prior to Stata 14
    real scalar p, p1, l
    
    p1 = strpos(s0, s)
    if (p1==0) return(0)
    l = strlen(s)
    if (l==0) return(1)
    p = p1 + l
    while (1) {
        p1 = strpos(substr(s0, p, .), s)
        if (p1==0) return(p-l)
        p = p + (p1-1) + l
    }
}

void texdoc_do_doblock(real scalar texmode, real scalar fh)
{
    real scalar      nodo, nodo0, cmdl, cmdl1, cmdlog0, noltrim, noltrim0,
                     offset, stlog, i, i0, j, j0, l, lb, cb, EOF
    string scalar    dobuf, cmdlogbuf, savecmd, line, tex1, tex2, texdocexit
    string rowvector s
    string colvector lines
    real colvector   dolines
    
    texdocexit = "// texdoc exit"
    tex1 = "/*tex"
    tex2 = "/***"
    EOF = 0
    nodo0 = (st_global("TeXdoc_nodo")!="")
    cmdlog0 = (st_global("TeXdoc_cmdlog")!="")
    noltrim0 = (st_global("TeXdoc_noltrim")!="")
    dobuf  = st_local("dobuf")  // buffer for stata commands
    unlink(dobuf)
    cmdlogbuf  = st_local("cmdlogbuf")  // buffer for cmdlog
    unlink(cmdlogbuf)
    savecmd = st_local("savecmd")
    fh = fopen(st_local("dofile"), "r")
    offset = strtoreal(st_local("offset"))
    fseek(fh, offset, -1)
    
    // find end of do block
    l = lb = cb = stlog = cmdl = cmdl1 = 0
    while ((line = fget(fh))!=J(0,0,"")) {
        if (lb==0) {                                // new command line
            cmdl1 = l   // ending line of prev command
            s = subinstr(line, char(9), " ")
            if (stlog==0) {                         // end of do block
                if (strrtrim(substr(strltrim(s),1,6))==tex1) {
                    texmode = 1
                    st_local("textag", "tex")
                    break
                }
                if (strrtrim(substr(strltrim(s),1,5))==tex2) {
                    texmode = 1
                    st_local("textag", "**")
                    break
                }
            }
            if (strtrim(stritrim(s))==texdocexit) {          // end of input
                EOF = 1
                break
            }
            s = texdoc_strip_comments(s, lb, cb)
        }
        else {                                      // continued command line
            s = s + texdoc_strip_comments(subinstr(line, char(9), " "), lb, cb)
        }
        if (lb) {                                   // command not complete yet
            l++
            continue
        }
        l++
        s = strtrim(stritrim(s))
        if (substr(s, 1, 7)=="texdoc ") {           // parse texdoc command
            s = tokens(substr(s, 8, .), ", ")
            if (stlog) {
                if (texdoc_cmdmatch(s, ("stlog", "close"), (1, 1))) {
                    stlog = 0
                    if (cmdl) break
                }
            }
            else if (texdoc_cmdmatch(s, "stlog", 1)) {
                stlog = 1
                for (j=2; j<=length(s); j++) {      // parse stlog using
                    if (s[j]=="using") {
                        stlog = 0
                        break
                    }
                    if (s[j]==",") break
                }
                if (stlog) {
                    cmdl = (cmdlog0 ? l + 1 : 0)    // parse [no]cmdlog option
                    for (; j<=length(s); j++) {
                        if (texdoc_cmdmatch(s[j], "cmdlog", 4)) {
                            cmdl = l + 1
                            break // cmdline takes precedence 
                        }
                        if (texdoc_cmdmatch(s[j], "nocmdlog", 6)) cmdl = 0 
                    }
                }
            }
            else if (texdoc_cmdmatch(s, "init", 1)) break
        }
        s = ""
    }
    if (line==J(0,0,"")) EOF = 1
    
    // read do block
    fseek(fh, offset, -1)
    lb = cb = stlog = nodo = noltrim = 0
    lines = J(l, 1, "")
    dolines = J(l, 1, 1)
    for (i=1; i<=l; i++) { 
        line = fget(fh)
        if (lb==0) {
            i0 = i; s = ""
        }
        s = s + texdoc_strip_comments(subinstr(line, char(9), " "), lb, cb)
        if (lb==0) {
            s = strtrim(stritrim(s))
            if (substr(s, 1, 7)=="texdoc ") {
                s = tokens(substr(s, 8, .), ", ")
                if (stlog) {                        // parse texdoc stlog close
                    if (texdoc_cmdmatch(s, ("stlog", "close"), (1, 1))) {
                        if (noltrim==0) _texdoc_do_blockltrim(lines, stlog+1, i0-1)
                        stlog = nodo = noltrim = 0
                        if (i0<i) dolines[|i0 \ i-1|] = J(i-i0, 1, 1)
                    }
                    if (nodo) dolines[i] = 0
                }                                   // parse texdoc stlog
                else if (texdoc_cmdmatch(s, "stlog", 1)) {
                    stlog = i
                    for (j=2; j<=length(s); j++) {  // parse stlog using
                        if (s[j]=="using") {
                            stlog = 0
                            break
                        }
                        if (s[j]==",") break
                    }
                    if (stlog) { // parse [no]do and [no]ltrim options
                        j0 = j
                        nodo = nodo0
                        for (; j<=length(s); j++) {
                            if (s[j]=="nodo") {
                                nodo = 1
                                break // nodo takes precedence 
                            }
                            if (s[j]=="do") nodo = 0
                        }
                        noltrim = noltrim0
                        for (j = j0; j<=length(s); j++) {
                            if (s[j]=="noltrim") {
                                noltrim = 1
                                break // noltrim takes precedence 
                            }
                            if (s[j]=="ltrim") noltrim = 0
                        }
                    }
                }
            }
            else if (nodo) dolines[i] = 0
        }
        else if (nodo) dolines[i] = 0
        lines[i] = line
    }
    if (EOF) st_local("offset", ".")
    else     st_local("offset", strofreal(ftell(fh)))
    fclose(fh)
    
    // write lines to do buffer
    if (l>0) {
        if (savecmd!="") texdoc_fput(savecmd, fh, lines, "a")
        if (cmdl) {
            if (cmdl1<cmdl) texdoc_fput(cmdlogbuf, fh, J(0, 1, ""), "w")
            else texdoc_fput(cmdlogbuf, fh, lines[|cmdl \ cmdl1|], "w")
        }
        texdoc_fput(dobuf, fh, select(lines, dolines), "w")
        if (lines!="") { // (do not process single empty line)
            st_local("flushdobuf","1")
        }
    }
}
void _texdoc_do_blockltrim(string colvector s0, real scalar a, real scalar b)
{
    real scalar l
    string colvector s
    
    if (a>b) return
    s = s0[|a \ b|]
    l = min((strlen(s) :- strlen(strltrim(s))))
    if (l<1) return
    st_global("TeXdoc_stindent", strofreal(l))
    s0[|a \ b|] = substr(s, l+1, .)
}

real scalar texdoc_cmdmatch(string rowvector s, string rowvector m, ///
    real rowvector l)
{   // checks whether the words in s match m, where l specifies the minimum 
    // abbreviation
    real scalar i, c

    c = length(s)
    for (i=1; i<=length(m); i++) {
        if (c<i) return(0)
        if (s[i]!=substr(m[i], 1, max((l[i], strlen(s[i]))))) return(0)
    }
    return(1)
}

// texdoc_strip_comments(): preprocessor for stata commandline
// - strips all comments from s0 ("* ...", "/* ... */", "//...", "///...")
//   taking account of quotes ("...", `"..."')
// - indicates whether line ends in a line break
// - indicates the inline comment nesting level at end of line (needed when 
//   reading the next line)
string scalar texdoc_strip_comments(
    string scalar s0, 
    real scalar lb, 
    real scalar cb)
{
    real scalar     i, wd, a, b, skip, first, cdq, dq
    string scalar   s, c0, c1, c, lq, rq
    // cdq: compound double quote nesting level
    // lq:  left compound double quote
    // rq:  right compound double quote
    // dq:  double quote indicator
    // lb:  end of line comment or line break indicator     (will be replaced)
    // cb:  inline comment nesting level                    (will be replaced)
    
    wd = strlen(s0)
    if (cb) a = 0
    else    a = 1
    b = wd = strlen(s0); first = 1
    skip = cdq = dq = lb = 0
    if (cb<0 | cb>=.) cb = 0
    s = ""; lq = "`" + `"""'; rq = `"""' + "'"
    c1 = " " // add a blank so that line can start with "*", "//", "///"
    for (i=1; i<=wd; i++) {
        c0 = c1
        c1 = substr(s0, i, 1)
        if (skip) {                     // skip one character
            skip = 0
            continue
        }
        if (first) first = (c0==" ")    // set to zero at first nonblank char
        c = c0+c1
        if (cdq) {                      // within compund double quotes
            if (c==lq) {
                cdq++; skip=1
            }
            else if (c==rq) {
                cdq--; skip=1
            }
        }
        else if (dq) {                  // within double quotes
            if (c0==`"""') dq=0
        }
        else if (cb) {                  // within inline comment
            if (c=="/*") {
                cb++; skip=1
            }
            else if (c=="*/") {
                cb--; skip=1
            }
            if (cb==0) a = i+1
        }
        else {
            if (lb) {                   // check for end of line comment
                if (c=="//") {
                    lb++
                    if (lb==3) {        // line stops with "///" or "///..."
                        b = i-3         // leading blank remains
                        break
                    }
                    if (i==wd) {        // line stops with "//"
                        b = i-2         // leading blank remains
                        break
                    }
                    c0 = c1
                    continue
                }
                else if (lb==2) {       // line stops with "//..."
                    b = i-3             // leading blank remains
                    break
                }
                else lb = 0
            }
            if (c==lq) {                // compound double quote start
                cdq++; skip=1
            }
            else if (c0==`"""') dq=1    // double quote start 
            else if (c=="/*") {         // inline comment start
                s = s + substr(s0, a, i-1-a)
                cb++; skip=1; a=0
            }
            else if (c==" /") lb++      // possible end of line comment
            else if (first) {           // * comment line
                if (c1=="*") {
                    b = i-1
                    break
                }
            }
        }
    }
    lb = (cb>0 | lb==3)                 // set line break indicator
    if (a>0) return(s + substr(s0, a, b-a+1))
    return(s)
}

// texdoc_strip_oom(): removes "texdoc stlog oom" from multiline command in 
// dofile or logfile (taking account of abbreviation, inline comments and line 
// breaks) and identifies the lines to be discarded
// modifies f and p; also resets b if log is specified
void texdoc_strip_oom(string colvector f, real colvector p,
    real scalar a, real scalar b, real scalar lskip, real scalar log)
{
    real scalar     i, wd, j, cb, lb, oom, gap
    string scalar   s, c0, c1, c, c00

    cb = 0
    j = a-1; i = 0; wd = 0; oom = 0; gap = 0; lb = 0; c00 = " "
    while (1) {
        if (++i>wd) {
            if (++j>b) break
            if (j>a) {
                if (lb==0 & cb==0) return // (should not happen)
            }
            i = 1
            if (log) s = substr(f[j], 3, .)
            else     s = subinstr(f[j], char(9), " ")
            wd = strlen(s)
            lb = 0
            c1 = " " // add a blank so that line can start with "//..."
        }
        c0 = c1
        c1 = substr(s, i, 1)
        c = c0 + c1
        if (cb) {                       // within inline comment
            if (c=="/*") {
                cb++; c1 = ""
            }
            else if (c=="*/") {
                cb--; c1 = ""
            }
        }
        else {
            if (lb) {
                if (c=="//") {          // end-of-line comment
                    i = wd; c1 = ""; continue
                }
                else lb = 0
            }
            if (c=="/*") {              // inline comment starts
                cb++; c1 = ""; continue
            }
            if (c==" /") {              // possible end-of-line comment
                lb = 1; continue
            }
            if (c1=="/") {              // inline comment
                if (substr(s, i+1, 1)=="*") continue
            }
            if (oom==0) {
                if (c0!="") c00 = c0
                if ((c00+c1)==" o") oom = 1 
            }
            else {
                if (gap==0) {
                    if (c1==substr("oom", oom+1, 1)) {
                        oom++; continue
                    }
                    if (c1!=" ") return     // (should not happen)
                }
                gap++
                if (gap==2) break
            }
        }
    }
    if (j>b) {      // (should not happen)
        j--; i = .
    }
    if (j>a) p[|a \ j-1|] = 0
    if (log) {
        f[j] = ". " + " " * lskip + substr(substr(f[j], 3, .), i, .)
        b = j       // reset b to first line command line
    }
    else     f[j] = " " * lskip + substr(f[j], i, .)
}

// texdoc_striplog_lbpos(): finds the byte position of line break in string 
// (" ///...") taking account of inline comments and quotes; does not take 
// account of "*..." and "//...", i.e. must check look for line break
// using texdoc_strip_comments() before applying texdoc_striplog_lbpos()
real scalar texdoc_striplog_lbpos(string scalar s, real scalar cb0)
{
    real scalar     i, wd, skip, cdq, dq, lb, lbpos, cb
    string scalar   c0, c1, c, lq, rq
    
    lbpos = 0
    wd = strlen(s)
    cb = cb0
    skip = cdq = dq = lb = 0
    lq = "`" + `"""'; rq = `"""' + "'"
    c1 = " " // add a blank so that line can start with "///"
    for (i=1; i<=wd; i++) {
        c0 = c1
        c1 = substr(s, i, 1)
        if (skip) {                     // skip one character
            skip = 0
            continue
        }
        c = c0+c1
        if (cdq) {                      // within compund double quotes
            if (c==lq) {
                cdq++; skip=1
            }
            else if (c==rq) {
                cdq--; skip=1
            }
        }
        else if (dq) {                  // within double quotes
            if (c0==`"""') dq=0
        }
        else if (cb) {                  // within inline comment
            if (c=="/*") { // take care of "/*/"
                cb++; skip=1
            }
            else if (c=="*/") { // take care of "*/*"
                cb--; skip=1
            }
        }
        else {
            if (lb) {                   // check for end of line comment
                if (c=="//") {
                    lb++
                    if (lb==3) {        // line stops with "///" or "///..."
                        lbpos = max((i-3,1))
                        break
                    }
                    continue
                }
                else lb = 0
            }
            if (c==lq) {                // compound double quote start
                cdq++; skip=1
            }
            else if (c0==`"""') dq=1    // double quote start 
            else if (c=="/*") {         // inline comment start
                cb++; skip=1
            }
            else if (c==" /") lb++      // possible end of line comment
        }
    }
    return(lbpos)
}

// texdoc_cat(): read file; simplified cat() with file handle argument
string colvector texdoc_cat(string scalar filename, real scalar fh)
{
        real scalar             i, n
        string matrix           EOF
        string colvector        res
        string scalar           line

        EOF = J(0, 0, "")
        fh  = fopen(filename, "r")
        // count lines
        i = 0
        while (1) {
            if (fget(fh)==EOF) break
            i++ 
        }
        res = J(n = i, 1, "")
        // read file
        fseek(fh, 0, -1)
        for (i=1; i<=n; i++) {
                if ((line=fget(fh))==EOF) {
                        /* unexpected EOF -- file must have changed */
                        fclose(fh)
                        if (--i) return(res[|1\i|])
                        return(J(0,1,""))
                }
                res[i] = line
        }
        fclose(fh)
        return(res)
}

// texdoc_fput(): write multiple lines to file
void texdoc_fput(string scalar fn, real scalar fh, string colvector s, 
    string scalar mode)
{
    real scalar i
    
    fh = fopen(fn, mode)
    for (i=1; i<=rows(s); i++) {
        fput(fh, s[i])
    }
    fclose(fh)
}

// texdoc_add_suffix(): add default suffix to filename if it has no suffix
void texdoc_add_suffix(string scalar macname, string scalar suffix)
{
    string scalar fn
    
    fn = st_local(macname)
    if (pathsuffix(fn)=="") fn = fn + suffix
    st_local(macname, fn)
}

// texdoc_add_abspath(): add absolute path if path is relative
void texdoc_add_abspath(string scalar macname)
{
    string scalar fn
    
    fn = st_local(macname)
    if (pathisabs(fn)==0) fn = pathjoin(pwd(), fn)
    st_local(macname, fn)
}

// texdoc_get_path(): returns path from filename in local cd
void texdoc_get_path(string scalar fn)
{
    string scalar path, basename
    
    pathsplit(fn, path, basename)
    st_local("path", path)
}

// texdoc_instance_fh(): creates a Mata global to be used as file handle later
// on; based on suggestion by W. Gould; allows closing open file handles on
// error or break
void texdoc_instance_fh(string scalar macname)
{
    real scalar   i
    string scalar fullname
    
    for (i=1; 1; i++) {
        fullname = sprintf("%s%g", "texdoc_fh_", i)
        if (crexternal(fullname) != NULL) {
            st_local(macname, fullname)
            return
        }
    }
}

// texdoc_closeout_fh(): close file handle if existing
void texdoc_closeout_fh(real scalar fh)
{
    if (fh!=.) (void) _fclose(fh)
}

end
